@@ -76,7 +76,7 @@ ngx_conf_parse(ngx_conf_t *cf, ngx_str_t *filename)
 
         /* open configuration file */
 
-        fd = ngx_open_file(filename->data, NGX_FILE_RDONLY, NGX_FILE_OPEN);
+        fd = ngx_open_file(filename->data, NGX_FILE_RDONLY, NGX_FILE_OPEN, 0);
         if (fd == NGX_INVALID_FILE) {
             ngx_conf_log_error(NGX_LOG_EMERG, cf, ngx_errno,
                                ngx_open_file_n " \"%s\" failed",
@@ -141,24 +141,26 @@ ngx_conf_parse(ngx_conf_t *cf, ngx_str_t *filename)
          */
 
         if (rc == NGX_ERROR) {
-            break;
+            goto done;
         }
 
         if (rc == NGX_CONF_BLOCK_DONE) {
+            if (!block) {
+                ngx_conf_log_error(NGX_LOG_EMERG, cf, 0, "unexpected \"}\"");
+                goto failed;
+            }
+
             block = 0;
         }
 
         if (rc == NGX_CONF_FILE_DONE && block) {
-            ngx_log_error(NGX_LOG_EMERG, cf->log, 0,
-                         "unexpected end of file in %s:%ui, expecting \"}\"",
-                         cf->conf_file->file.name.data,
-                         cf->conf_file->line);
-            rc = NGX_ERROR;
-            break;
+            ngx_conf_log_error(NGX_LOG_EMERG, cf, 0,
+                               "unexpected end of file, expecting \"}\"");
+            goto failed;
         }
 
         if (rc != NGX_OK && rc != NGX_CONF_BLOCK_START) {
-            break;
+            goto done;
         }
 
         if (cf->handler) {
@@ -174,26 +176,27 @@ ngx_conf_parse(ngx_conf_t *cf, ngx_str_t *filename)
             }
 
             if (rv == NGX_CONF_ERROR) {
-                rc = NGX_ERROR;
-                break;
+                goto failed;
             }
 
-            ngx_log_error(NGX_LOG_EMERG, cf->log, 0,
-                         "%s in %s:%ui",
-                         rv, cf->conf_file->file.name.data,
-                         cf->conf_file->line);
-            rc = NGX_ERROR;
-            break;
+            ngx_conf_log_error(NGX_LOG_EMERG, cf, 0, rv);
+
+            goto failed;
         }
 
 
         rc = ngx_conf_handler(cf, rc);
 
         if (rc == NGX_ERROR) {
-            break;
+            goto failed;
         }
     }
 
+failed:
+
+    rc = NGX_ERROR;
+
+done:
 
     if (filename) {
         ngx_free(cf->conf_file->buffer->start);
@@ -267,20 +270,16 @@ ngx_conf_handler(ngx_conf_t *cf, ngx_int_t last)
             }
 
             if (!(cmd->type & NGX_CONF_BLOCK) && last != NGX_OK) {
-                ngx_log_error(NGX_LOG_EMERG, cf->log, 0,
-                              "directive \"%s\" in %s:%ui "
-                              "is not terminated by \";\"",
-                              name->data, cf->conf_file->file.name.data,
-                              cf->conf_file->line);
+                ngx_conf_log_error(NGX_LOG_EMERG, cf, 0,
+                                  "directive \"%s\" is not terminated by \";\"",
+                                  name->data);
                 return NGX_ERROR;
             }
 
             if ((cmd->type & NGX_CONF_BLOCK) && last != NGX_CONF_BLOCK_START) {
-                ngx_log_error(NGX_LOG_EMERG, cf->log, 0,
-                              "directive \"%s\" in %s:%ui "
-                              "has not the opening \"{\"",
-                              name->data, cf->conf_file->file.name.data,
-                              cf->conf_file->line);
+                ngx_conf_log_error(NGX_LOG_EMERG, cf, 0,
+                                   "directive \"%s\" has no opening \"{\"",
+                                   name->data);
                 return NGX_ERROR;
             }
 
@@ -344,40 +343,31 @@ ngx_conf_handler(ngx_conf_t *cf, ngx_int_t last)
                 return NGX_ERROR;
             }
 
-            ngx_log_error(NGX_LOG_EMERG, cf->log, 0,
-                          "the \"%s\" directive %s in %s:%ui",
-                          name->data, rv, cf->conf_file->file.name.data,
-                          cf->conf_file->line);
+            ngx_conf_log_error(NGX_LOG_EMERG, cf, 0,
+                               "\"%s\" directive %s", name->data, rv);
 
             return NGX_ERROR;
         }
     }
 
     if (multi == 0) {
-        ngx_log_error(NGX_LOG_EMERG, cf->log, 0,
-                      "unknown directive \"%s\" in %s:%ui",
-                      name->data, cf->conf_file->file.name.data,
-                      cf->conf_file->line);
+        ngx_conf_log_error(NGX_LOG_EMERG, cf, 0,
+                           "unknown directive \"%s\"", name->data);
 
         return NGX_ERROR;
     }
 
 not_allowed:
 
-    ngx_log_error(NGX_LOG_EMERG, cf->log, 0,
-                  "directive \"%s\" in %s:%ui "
-                  "is not allowed here",
-                  name->data, cf->conf_file->file.name.data,
-                  cf->conf_file->line);
+    ngx_conf_log_error(NGX_LOG_EMERG, cf, 0,
+                       "\"%s\" directive is not allowed here", name->data);
     return NGX_ERROR;
 
 invalid:
 
-    ngx_log_error(NGX_LOG_EMERG, cf->log, 0,
-                  "invalid number arguments in "
-                  "directive \"%s\" in %s:%ui",
-                  name->data, cf->conf_file->file.name.data,
-                  cf->conf_file->line);
+    ngx_conf_log_error(NGX_LOG_EMERG, cf, 0,
+                       "invalid number arguments in \"%s\" directive",
+                       name->data);
 
     return NGX_ERROR;
 }
@@ -412,11 +402,9 @@ ngx_conf_read_token(ngx_conf_t *cf)
                                  >= ngx_file_size(&cf->conf_file->file.info))
             {
                 if (cf->args->nelts > 0) {
-                    ngx_log_error(NGX_LOG_EMERG, cf->log, 0,
-                                  "unexpected end of file in %s:%ui, "
-                                  "expecting \";\" or \"}\"",
-                                  cf->conf_file->file.name.data,
-                                  cf->conf_file->line);
+                    ngx_conf_log_error(NGX_LOG_EMERG, cf, 0,
+                                  "unexpected end of file, "
+                                  "expecting \";\" or \"}\"");
                     return NGX_ERROR;
                 }
 
@@ -480,11 +468,8 @@ ngx_conf_read_token(ngx_conf_t *cf)
                 need_space = 0;
 
             } else {
-                 ngx_log_error(NGX_LOG_EMERG, cf->log, 0,
-                               "unexpected \"%c\" in %s:%ui",
-                               ch, cf->conf_file->file.name.data,
-                               cf->conf_file->line);
-
+                 ngx_conf_log_error(NGX_LOG_EMERG, cf, 0,
+                                    "unexpected \"%c\"", ch);
                  return NGX_ERROR;
             }
         }
@@ -501,10 +486,8 @@ ngx_conf_read_token(ngx_conf_t *cf)
             case ';':
             case '{':
                 if (cf->args->nelts == 0) {
-                    ngx_log_error(NGX_LOG_EMERG, cf->log, 0,
-                                  "unexpected \"%c\" in %s:%ui",
-                                  ch, cf->conf_file->file.name.data,
-                                  cf->conf_file->line);
+                    ngx_conf_log_error(NGX_LOG_EMERG, cf, 0,
+                                       "unexpected \"%c\"", ch);
                     return NGX_ERROR;
                 }
 
@@ -516,10 +499,8 @@ ngx_conf_read_token(ngx_conf_t *cf)
 
             case '}':
                 if (cf->args->nelts != 0) {
-                    ngx_log_error(NGX_LOG_EMERG, cf->log, 0,
-                                  "unexpected \"}\" in %s:%ui",
-                                  cf->conf_file->file.name.data,
-                                  cf->conf_file->line);
+                    ngx_conf_log_error(NGX_LOG_EMERG, cf, 0,
+                                       "unexpected \"}\"");
                     return NGX_ERROR;
                 }
 
@@ -883,10 +864,10 @@ ngx_conf_set_flag_slot(ngx_conf_t *cf, ngx_command_t *cmd, void *conf)
 
     value = cf->args->elts;
 
-    if (ngx_strcasecmp(value[1].data, "on") == 0) {
+    if (ngx_strcasecmp(value[1].data, (u_char *) "on") == 0) {
         *fp = 1;
 
-    } else if (ngx_strcasecmp(value[1].data, "off") == 0) {
+    } else if (ngx_strcasecmp(value[1].data, (u_char *) "off") == 0) {
         *fp = 0;
 
     } else {
