@@ -24,29 +24,12 @@ typedef struct {
 } ngx_keyval_t;
 
 
-#define ngx_string(str)  { sizeof(str) - 1, (u_char *) str }
-#define ngx_null_string  { 0, NULL }
+#define ngx_string(str)     { sizeof(str) - 1, (u_char *) str }
+#define ngx_null_string     { 0, NULL }
 
 
-#define ngx_tolower(c)     (u_char) ((c >= 'A' && c <= 'Z') ? (c | 0x20) : c)
-#define ngx_toupper(c)     (u_char) ((c >= 'a' && c <= 'z') ? (c & ~0x20) : c)
-
-
-#if (NGX_WIN32)
-
-#define ngx_strncasecmp(s1, s2, n)                                           \
-                           strnicmp((const char *) s1, (const char *) s2, n)
-#define ngx_strcasecmp(s1, s2)                                               \
-                           stricmp((const char *) s1, (const char *) s2)
-
-#else
-
-#define ngx_strncasecmp(s1, s2, n)                                           \
-                           strncasecmp((const char *) s1, (const char *) s2, n)
-#define ngx_strcasecmp(s1, s2)                                               \
-                           strcasecmp((const char *) s1, (const char *) s2)
-
-#endif
+#define ngx_tolower(c)      (u_char) ((c >= 'A' && c <= 'Z') ? (c | 0x20) : c)
+#define ngx_toupper(c)      (u_char) ((c >= 'a' && c <= 'z') ? (c & ~0x20) : c)
 
 
 #define ngx_strncmp(s1, s2, n)  strncmp((const char *) s1, (const char *) s2, n)
@@ -57,6 +40,7 @@ typedef struct {
 
 
 #define ngx_strstr(s1, s2)  strstr((const char *) s1, (const char *) s2)
+#define ngx_strchr(s1, c)   strchr((const char *) s1, (int) c)
 #define ngx_strlen(s)       strlen((const char *) s)
 
 
@@ -128,8 +112,12 @@ u_char * ngx_cdecl ngx_sprintf(u_char *buf, const char *fmt, ...);
 u_char * ngx_cdecl ngx_snprintf(u_char *buf, size_t max, const char *fmt, ...);
 u_char *ngx_vsnprintf(u_char *buf, size_t max, const char *fmt, va_list args);
 
+ngx_int_t ngx_strcasecmp(u_char *s1, u_char *s2);
+ngx_int_t ngx_strncasecmp(u_char *s1, u_char *s2, size_t n);
+
 ngx_int_t ngx_rstrncmp(u_char *s1, u_char *s2, size_t n);
 ngx_int_t ngx_rstrncasecmp(u_char *s1, u_char *s2, size_t n);
+ngx_int_t ngx_memn2cmp(u_char *s1, u_char *s2, size_t n1, size_t n2);
 
 ngx_int_t ngx_atoi(u_char *line, size_t n);
 ssize_t ngx_atosz(u_char *line, size_t n);
@@ -148,25 +136,27 @@ ngx_int_t ngx_decode_base64(ngx_str_t *dst, ngx_str_t *src);
 
 uint32_t ngx_utf_decode(u_char **p, size_t n);
 size_t ngx_utf_length(u_char *p, size_t n);
-u_char * ngx_utf_cpystrn(u_char *dst, u_char *src, size_t n);
+u_char *ngx_utf_cpystrn(u_char *dst, u_char *src, size_t n);
 
 
-#define NGX_ESCAPE_URI     0
-#define NGX_ESCAPE_ARGS    1
-#define NGX_ESCAPE_HTML    2
+#define NGX_ESCAPE_URI       0
+#define NGX_ESCAPE_ARGS      1
+#define NGX_ESCAPE_HTML      2
 
-#define NGX_UNESCAPE_URI   1
+#define NGX_UNESCAPE_URI     1
 
 uintptr_t ngx_escape_uri(u_char *dst, u_char *src, size_t size,
     ngx_uint_t type);
 void ngx_unescape_uri(u_char **dst, u_char **src, size_t size, ngx_uint_t type);
 
 
-#define  ngx_qsort                qsort
+void ngx_sort(void *base, size_t n, size_t size,
+    int (*cmp)(const void *, const void *));
+#define ngx_qsort            qsort
 
 
-#define  ngx_value_helper(n)      #n
-#define  ngx_value(n)             ngx_value_helper(n)
+#define ngx_value_helper(n)  #n
+#define ngx_value(n)         ngx_value_helper(n)
 
 
 #endif /* _NGX_STRING_H_INCLUDED_ */
