@@ -107,8 +107,8 @@ static char *ngx_http_ssi_types(ngx_conf_t *cf, ngx_command_t *cmd, void *conf);
 static ngx_int_t ngx_http_ssi_preconfiguration(ngx_conf_t *cf);
 static void *ngx_http_ssi_create_main_conf(ngx_conf_t *cf);
 static char *ngx_http_ssi_init_main_conf(ngx_conf_t *cf, void *conf);
-static void *ngx_http_ssi_create_conf(ngx_conf_t *cf);
-static char *ngx_http_ssi_merge_conf(ngx_conf_t *cf,
+static void *ngx_http_ssi_create_loc_conf(ngx_conf_t *cf);
+static char *ngx_http_ssi_merge_loc_conf(ngx_conf_t *cf,
     void *parent, void *child);
 static ngx_int_t ngx_http_ssi_filter_init(ngx_conf_t *cf);
 
@@ -173,8 +173,8 @@ static ngx_http_module_t  ngx_http_ssi_filter_module_ctx = {
     NULL,                                  /* create server configuration */
     NULL,                                  /* merge server configuration */
 
-    ngx_http_ssi_create_conf,              /* create location configuration */
-    ngx_http_ssi_merge_conf                /* merge location configuration */
+    ngx_http_ssi_create_loc_conf,          /* create location configuration */
+    ngx_http_ssi_merge_loc_conf            /* merge location configuration */
 };
 
 
@@ -1020,6 +1020,7 @@ ngx_http_ssi_parse(ngx_http_request_t *r, ngx_http_ssi_ctx_t *ctx)
                 ch = *p;
             }
 
+            ctx->state = state;
             ctx->pos = p;
             ctx->looked = looked;
             ctx->copy_end = p;
@@ -1890,10 +1891,13 @@ ngx_http_ssi_include(ngx_http_request_t *r, ngx_http_ssi_ctx_t *ctx,
             return NGX_HTTP_SSI_ERROR;
         }
 
-        if (wait->len == 2 && ngx_strncasecmp(wait->data, "no", 2) == 0) {
+        if (wait->len == 2
+            && ngx_strncasecmp(wait->data, (u_char *) "no", 2) == 0)
+        {
             wait = NULL;
 
-        } else if (wait->len != 3 || ngx_strncasecmp(wait->data, "yes", 3) != 0)
+        } else if (wait->len != 3
+                   || ngx_strncasecmp(wait->data, (u_char *) "yes", 3) != 0)
         {
             ngx_log_error(NGX_LOG_ERR, r->connection->log, 0,
                           "invalid value \"%V\" in the \"wait\" parameter",
@@ -2373,7 +2377,7 @@ ngx_http_ssi_if(ngx_http_request_t *r, ngx_http_ssi_ctx_t *ctx,
         p++;
     }
 
-    if (p < last && *p == '/') {
+    if (p < last - 1 && *p == '/') {
         if (*(last - 1) != '/') {
             goto invalid_expression;
         }
@@ -2381,7 +2385,7 @@ ngx_http_ssi_if(ngx_http_request_t *r, ngx_http_ssi_ctx_t *ctx,
         noregex = 0;
         flags = NGX_HTTP_SSI_ADD_ZERO;
         last--;
-	p++;
+        p++;
 
     } else {
         noregex = 1;
@@ -2743,7 +2747,7 @@ ngx_http_ssi_init_main_conf(ngx_conf_t *cf, void *conf)
 
 
 static void *
-ngx_http_ssi_create_conf(ngx_conf_t *cf)
+ngx_http_ssi_create_loc_conf(ngx_conf_t *cf)
 {
     ngx_http_ssi_loc_conf_t  *slcf;
 
@@ -2770,7 +2774,7 @@ ngx_http_ssi_create_conf(ngx_conf_t *cf)
 
 
 static char *
-ngx_http_ssi_merge_conf(ngx_conf_t *cf, void *parent, void *child)
+ngx_http_ssi_merge_loc_conf(ngx_conf_t *cf, void *parent, void *child)
 {
     ngx_http_ssi_loc_conf_t *prev = parent;
     ngx_http_ssi_loc_conf_t *conf = child;
