@@ -11,12 +11,8 @@
 
 
 typedef struct {
-    PerlInterpreter  **free_perls;
-    ngx_uint_t         interp;
-    ngx_uint_t         nalloc;
-    ngx_uint_t         interp_max;
-
     PerlInterpreter   *perl;
+    HV                *nginx;
     ngx_str_t          modules;
     ngx_array_t        requires;
 } ngx_http_perl_main_conf_t;
@@ -45,21 +41,15 @@ static ngx_int_t ngx_http_perl_ssi(ngx_http_request_t *r,
     ngx_http_ssi_ctx_t *ssi_ctx, ngx_str_t **params);
 #endif
 
-static ngx_int_t
-    ngx_http_perl_get_interpreter(ngx_http_perl_main_conf_t *pmcf,
-    PerlInterpreter **perl, ngx_log_t *log);
-static ngx_inline void
-    ngx_http_perl_free_interpreter(ngx_http_perl_main_conf_t *pmcf,
-    PerlInterpreter *perl);
+static void ngx_http_perl_sleep_handler(ngx_http_request_t *r);
 static char *ngx_http_perl_init_interpreter(ngx_conf_t *cf,
     ngx_http_perl_main_conf_t *pmcf);
-static PerlInterpreter *
-    ngx_http_perl_create_interpreter(ngx_http_perl_main_conf_t *pmcf,
-    ngx_log_t *log);
+static PerlInterpreter *ngx_http_perl_create_interpreter(ngx_conf_t *cf,
+    ngx_http_perl_main_conf_t *pmcf);
 static ngx_int_t ngx_http_perl_run_requires(pTHX_ ngx_array_t *requires,
     ngx_log_t *log);
 static ngx_int_t ngx_http_perl_call_handler(pTHX_ ngx_http_request_t *r,
-    SV *sub, ngx_str_t **args, ngx_str_t *handler, ngx_str_t *rv);
+    HV *nginx, SV *sub, ngx_str_t **args, ngx_str_t *handler, ngx_str_t *rv);
 static void ngx_http_perl_eval_anon_sub(pTHX_ ngx_str_t *handler, SV **sv);
 
 static ngx_int_t ngx_http_perl_preconfiguration(ngx_conf_t *cf);
@@ -72,16 +62,12 @@ static char *ngx_http_perl_require(ngx_conf_t *cf, ngx_command_t *cmd,
     void *conf);
 static char *ngx_http_perl(ngx_conf_t *cf, ngx_command_t *cmd, void *conf);
 static char *ngx_http_perl_set(ngx_conf_t *cf, ngx_command_t *cmd, void *conf);
-static char *ngx_http_perl_interp_max_unsupported(ngx_conf_t *cf, void *post,
-    void *data);
-#if (NGX_HAVE_PERL_CLONE || NGX_HAVE_PERL_MULTIPLICITY)
+
+#if (NGX_HAVE_PERL_MULTIPLICITY)
 static void ngx_http_perl_cleanup_perl(void *data);
 #endif
-static void ngx_http_perl_cleanup_sv(void *data);
 
-
-static ngx_conf_post_handler_pt  ngx_http_perl_interp_max_p =
-    ngx_http_perl_interp_max_unsupported;
+static void ngx_http_perl_exit(ngx_cycle_t *cycle);
 
 
 static ngx_command_t  ngx_http_perl_commands[] = {
@@ -100,15 +86,8 @@ static ngx_command_t  ngx_http_perl_commands[] = {
       0,
       NULL },
 
-    { ngx_string("perl_interp_max"),
-      NGX_HTTP_MAIN_CONF|NGX_CONF_TAKE1,
-      ngx_conf_set_num_slot,
-      NGX_HTTP_MAIN_CONF_OFFSET,
-      offsetof(ngx_http_perl_main_conf_t, interp_max),
-      &ngx_http_perl_interp_max_p },
-
     { ngx_string("perl"),
-      NGX_HTTP_LOC_CONF|NGX_CONF_TAKE1,
+      NGX_HTTP_LOC_CONF|NGX_HTTP_LMT_CONF|NGX_CONF_TAKE1,
       ngx_http_perl,
       NGX_HTTP_LOC_CONF_OFFSET,
       0,
@@ -151,7 +130,7 @@ ngx_module_t  ngx_http_perl_module = {
     NULL,                                  /* init thread */
     NULL,                                  /* exit thread */
     NULL,                                  /* exit process */
-    NULL,                                  /* exit master */
+    ngx_http_perl_exit,                    /* exit master */
     NGX_MODULE_V1_PADDING
 };
 
@@ -200,24 +179,6 @@ ngx_http_perl_handler(ngx_http_request_t *r)
     ngx_http_perl_handle_request(r);
 
     return NGX_DONE;
-
-#if 0
-    r->request_body_in_single_buf = 1;
-    r->request_body_in_persistent_file = 1;
-    r->request_body_delete_incomplete_file = 1;
-
-    if (r->request_body_in_file_only) {
-        r->request_body_file_log_level = 0;
-    }
-
-    rc = ngx_http_read_client_request_body(r, ngx_http_perl_handle_request);
-
-    if (rc >= NGX_HTTP_SPECIAL_RESPONSE) {
-        return rc;
-    }
-
-    return NGX_DONE;
-#endif
 }
 
 
@@ -239,7 +200,7 @@ ngx_http_perl_handle_request(ngx_http_request_t *r)
         ctx = ngx_pcalloc(r->pool, sizeof(ngx_http_perl_ctx_t));
         if (ctx == NULL) {
             ngx_http_finalize_request(r, NGX_ERROR);
-	    return;
+            return;
         }
 
         ngx_http_set_ctx(r, ctx, ngx_http_perl_module);
@@ -247,16 +208,10 @@ ngx_http_perl_handle_request(ngx_http_request_t *r)
 
     pmcf = ngx_http_get_module_main_conf(r, ngx_http_perl_module);
 
-    rc = ngx_http_perl_get_interpreter(pmcf, &ctx->perl, r->connection->log);
-
-    if (rc != NGX_OK) {
-        ngx_http_finalize_request(r, rc);
-        return;
-    }
-
     {
 
-    dTHXa(ctx->perl);
+    dTHXa(pmcf->perl);
+    PERL_SET_CONTEXT(pmcf->perl);
 
     if (ctx->next == NULL) {
         plcf = ngx_http_get_module_loc_conf(r, ngx_http_perl_module);
@@ -269,12 +224,11 @@ ngx_http_perl_handle_request(ngx_http_request_t *r)
         ctx->next = NULL;
     }
 
-    rc = ngx_http_perl_call_handler(aTHX_ r, sub, NULL, handler, NULL);
+    rc = ngx_http_perl_call_handler(aTHX_ r, pmcf->nginx, sub, NULL, handler,
+                                    NULL);
 
     }
 
-    ngx_http_perl_free_interpreter(pmcf, ctx->perl);
-
     if (rc > 600) {
         rc = NGX_OK;
     }
@@ -293,6 +247,12 @@ ngx_http_perl_handle_request(ngx_http_request_t *r)
     ctx->filename.data = NULL;
     ctx->redirect_uri.len = 0;
 
+    if (ctx->sleep) {
+        ngx_add_timer(r->connection->write, (ngx_msec_t) ctx->sleep);
+        r->write_event_handler = ngx_http_perl_sleep_handler;
+        ctx->sleep = 0;
+    }
+
     if (ctx->done || ctx->next) {
         return;
     }
@@ -311,6 +271,28 @@ ngx_http_perl_handle_request(ngx_http_request_t *r)
 }
 
 
+static void
+ngx_http_perl_sleep_handler(ngx_http_request_t *r)
+{
+    ngx_event_t  *wev;
+
+    ngx_log_debug0(NGX_LOG_DEBUG_HTTP, r->connection->log, 0,
+                   "perl sleep handler");
+
+    wev = r->connection->write;
+
+    if (wev->timedout) {
+        wev->timedout = 0;
+        ngx_http_perl_handle_request(r);
+        return;
+    }
+
+    if (ngx_handle_write_event(wev, 0) == NGX_ERROR) {
+        ngx_http_finalize_request(r, NGX_HTTP_INTERNAL_SERVER_ERROR);
+    }
+}
+
+
 static ngx_int_t
 ngx_http_perl_variable(ngx_http_request_t *r, ngx_http_variable_value_t *v,
     uintptr_t data)
@@ -338,25 +320,18 @@ ngx_http_perl_variable(ngx_http_request_t *r, ngx_http_variable_value_t *v,
 
     pmcf = ngx_http_get_module_main_conf(r, ngx_http_perl_module);
 
-    rc = ngx_http_perl_get_interpreter(pmcf, &ctx->perl, r->connection->log);
-
-    if (rc != NGX_OK) {
-        return rc;
-    }
-
     value.data = NULL;
 
     {
 
-    dTHXa(ctx->perl);
+    dTHXa(pmcf->perl);
+    PERL_SET_CONTEXT(pmcf->perl);
 
-    rc = ngx_http_perl_call_handler(aTHX_ r, pv->sub, NULL,
+    rc = ngx_http_perl_call_handler(aTHX_ r, pmcf->nginx, pv->sub, NULL,
                                     &pv->handler, &value);
 
     }
 
-    ngx_http_perl_free_interpreter(pmcf, ctx->perl);
-
     if (value.data) {
         v->len = value.len;
         v->valid = 1;
@@ -371,6 +346,9 @@ ngx_http_perl_variable(ngx_http_request_t *r, ngx_http_variable_value_t *v,
     ctx->filename.data = NULL;
     ctx->redirect_uri.len = 0;
 
+    ngx_log_debug0(NGX_LOG_DEBUG_HTTP, r->connection->log, 0,
+                   "perl variable done");
+
     return rc;
 }
 
@@ -390,8 +368,6 @@ ngx_http_perl_ssi(ngx_http_request_t *r, ngx_http_ssi_ctx_t *ssi_ctx,
     ngx_log_debug0(NGX_LOG_DEBUG_HTTP, r->connection->log, 0,
                    "perl ssi handler");
 
-    pmcf = ngx_http_get_module_main_conf(r, ngx_http_perl_module);
-
     ctx = ngx_http_get_module_ctx(r, ngx_http_perl_module);
 
     if (ctx == NULL) {
@@ -403,11 +379,7 @@ ngx_http_perl_ssi(ngx_http_request_t *r, ngx_http_ssi_ctx_t *ssi_ctx,
         ngx_http_set_ctx(r, ctx, ngx_http_perl_module);
     }
 
-    rc = ngx_http_perl_get_interpreter(pmcf, &ctx->perl, r->connection->log);
-
-    if (rc != NGX_OK) {
-        return rc;
-    }
+    pmcf = ngx_http_get_module_main_conf(r, ngx_http_perl_module);
 
     ctx->ssi = ssi_ctx;
 
@@ -416,7 +388,8 @@ ngx_http_perl_ssi(ngx_http_request_t *r, ngx_http_ssi_ctx_t *ssi_ctx,
 
     {
 
-    dTHXa(ctx->perl);
+    dTHXa(pmcf->perl);
+    PERL_SET_CONTEXT(pmcf->perl);
 
 #if 0
 
@@ -438,65 +411,30 @@ ngx_http_perl_ssi(ngx_http_request_t *r, ngx_http_ssi_ctx_t *ssi_ctx,
 
     sv = newSVpvn((char *) handler->data, handler->len);
 
-    rc = ngx_http_perl_call_handler(aTHX_ r, sv, &params[NGX_HTTP_PERL_SSI_ARG],
+    rc = ngx_http_perl_call_handler(aTHX_ r, pmcf->nginx, sv,
+                                    &params[NGX_HTTP_PERL_SSI_ARG],
                                     handler, NULL);
 
     SvREFCNT_dec(sv);
 
     }
 
-    ngx_http_perl_free_interpreter(pmcf, ctx->perl);
-
     ctx->filename.data = NULL;
     ctx->redirect_uri.len = 0;
     ctx->ssi = NULL;
 
+    ngx_log_debug0(NGX_LOG_DEBUG_HTTP, r->connection->log, 0, "perl ssi done");
+
     return rc;
 }
 
 #endif
 
 
-static ngx_int_t
-ngx_http_perl_get_interpreter(ngx_http_perl_main_conf_t *pmcf,
-    PerlInterpreter **perl, ngx_log_t *log)
-{
-    if (pmcf->interp) {
-        pmcf->interp--;
-
-        *perl = pmcf->free_perls[pmcf->interp];
-
-        return NGX_OK;
-    }
-
-    if (pmcf->nalloc < pmcf->interp_max) {
-        *perl = ngx_http_perl_create_interpreter(pmcf, log);
-
-        if (*perl) {
-            return NGX_OK;
-        }
-
-        return NGX_HTTP_INTERNAL_SERVER_ERROR;
-    }
-
-    ngx_log_error(NGX_LOG_ALERT, log, 0, "no free perl interpreter");
-
-    return NGX_HTTP_SERVICE_UNAVAILABLE;
-}
-
-
-static ngx_inline void
-ngx_http_perl_free_interpreter(ngx_http_perl_main_conf_t *pmcf,
-    PerlInterpreter *perl)
-{
-    pmcf->free_perls[pmcf->interp++] = perl;
-}
-
-
 static char *
 ngx_http_perl_init_interpreter(ngx_conf_t *cf, ngx_http_perl_main_conf_t *pmcf)
 {
-#if (NGX_HAVE_PERL_CLONE || NGX_HAVE_PERL_MULTIPLICITY)
+#if (NGX_HAVE_PERL_MULTIPLICITY)
     ngx_pool_cleanup_t       *cln;
 
     cln = ngx_pool_cleanup_add(cf->pool, 0);
@@ -520,9 +458,14 @@ ngx_http_perl_init_interpreter(ngx_conf_t *cf, ngx_http_perl_main_conf_t *pmcf)
         }
     }
 
-#if !(NGX_HAVE_PERL_CLONE || NGX_HAVE_PERL_MULTIPLICITY)
+#if !(NGX_HAVE_PERL_MULTIPLICITY)
 
     if (perl) {
+
+        if (ngx_set_environment(cf->cycle, NULL) == NULL) {
+            return NGX_CONF_ERROR;
+        }
+
         if (ngx_http_perl_run_requires(aTHX_ &pmcf->requires, cf->log)
             != NGX_OK)
         {
@@ -530,22 +473,26 @@ ngx_http_perl_init_interpreter(ngx_conf_t *cf, ngx_http_perl_main_conf_t *pmcf)
         }
 
         pmcf->perl = perl;
+        pmcf->nginx = nginx_stash;
 
         return NGX_CONF_OK;
     }
 
 #endif
 
-    PERL_SYS_INIT(&ngx_argc, &ngx_argv);
+    if (nginx_stash == NULL) {
+        PERL_SYS_INIT(&ngx_argc, &ngx_argv);
+    }
 
-    pmcf->perl = ngx_http_perl_create_interpreter(pmcf, cf->log);
+    pmcf->perl = ngx_http_perl_create_interpreter(cf, pmcf);
 
     if (pmcf->perl == NULL) {
-        PERL_SYS_TERM();
         return NGX_CONF_ERROR;
     }
 
-#if (NGX_HAVE_PERL_CLONE || NGX_HAVE_PERL_MULTIPLICITY)
+    pmcf->nginx = nginx_stash;
+
+#if (NGX_HAVE_PERL_MULTIPLICITY)
 
     cln->handler = ngx_http_perl_cleanup_perl;
     cln->data = pmcf->perl;
@@ -561,8 +508,8 @@ ngx_http_perl_init_interpreter(ngx_conf_t *cf, ngx_http_perl_main_conf_t *pmcf)
 
 
 static PerlInterpreter *
-ngx_http_perl_create_interpreter(ngx_http_perl_main_conf_t *pmcf,
-    ngx_log_t *log)
+ngx_http_perl_create_interpreter(ngx_conf_t *cf,
+    ngx_http_perl_main_conf_t *pmcf)
 {
     int                n;
     STRLEN             len;
@@ -570,45 +517,24 @@ ngx_http_perl_create_interpreter(ngx_http_perl_main_conf_t *pmcf,
     char              *ver, *embedding[6];
     PerlInterpreter   *perl;
 
-    ngx_log_debug0(NGX_LOG_DEBUG_HTTP, log, 0, "create perl interpreter");
+    ngx_log_debug0(NGX_LOG_DEBUG_HTTP, cf->log, 0, "create perl interpreter");
 
-#if (NGX_HAVE_PERL_CLONE)
-
-    if (pmcf->perl) {
-
-        perl = perl_clone(pmcf->perl, CLONEf_KEEP_PTR_TABLE);
-        if (perl == NULL) {
-            ngx_log_error(NGX_LOG_ALERT, log, 0, "perl_clone() failed");
-            return NULL;
-        }
-
-        {
-
-        dTHXa(perl);
-
-        ptr_table_free(PL_ptr_table);
-        PL_ptr_table = NULL;
-
-        }
-
-        pmcf->nalloc++;
-
-        return perl;
-    }
-
-#endif
-
-    perl = perl_alloc();
-    if (perl == NULL) {
-        ngx_log_error(NGX_LOG_ALERT, log, 0, "perl_alloc() failed");
+    if (ngx_set_environment(cf->cycle, NULL) == NULL) {
         return NULL;
     }
 
-    perl_construct(perl);
+    perl = perl_alloc();
+    if (perl == NULL) {
+        ngx_log_error(NGX_LOG_ALERT, cf->log, 0, "perl_alloc() failed");
+        return NULL;
+    }
 
     {
 
     dTHXa(perl);
+    PERL_SET_CONTEXT(perl);
+
+    perl_construct(perl);
 
 #ifdef PERL_EXIT_DESTRUCT_END
     PL_exit_flags |= PERL_EXIT_DESTRUCT_END;
@@ -632,7 +558,7 @@ ngx_http_perl_create_interpreter(ngx_http_perl_main_conf_t *pmcf,
     n = perl_parse(perl, ngx_http_perl_xs_init, n, embedding, NULL);
 
     if (n != 0) {
-        ngx_log_error(NGX_LOG_ALERT, log, 0, "perl_parse() failed: %d", n);
+        ngx_log_error(NGX_LOG_ALERT, cf->log, 0, "perl_parse() failed: %d", n);
         goto fail;
     }
 
@@ -640,20 +566,18 @@ ngx_http_perl_create_interpreter(ngx_http_perl_main_conf_t *pmcf,
     ver = SvPV(sv, len);
 
     if (ngx_strcmp(ver, NGINX_VERSION) != 0) {
-        ngx_log_error(NGX_LOG_ALERT, log, 0,
+        ngx_log_error(NGX_LOG_ALERT, cf->log, 0,
                       "version " NGINX_VERSION " of nginx.pm is required, "
                       "but %s was found", ver);
         goto fail;
     }
 
-    if (ngx_http_perl_run_requires(aTHX_ &pmcf->requires, log) != NGX_OK) {
+    if (ngx_http_perl_run_requires(aTHX_ &pmcf->requires, cf->log) != NGX_OK) {
         goto fail;
     }
 
     }
 
-    pmcf->nalloc++;
-
     return perl;
 
 fail:
@@ -699,7 +623,7 @@ ngx_http_perl_run_requires(pTHX_ ngx_array_t *requires, ngx_log_t *log)
 
 
 static ngx_int_t
-ngx_http_perl_call_handler(pTHX_ ngx_http_request_t *r, SV *sub,
+ngx_http_perl_call_handler(pTHX_ ngx_http_request_t *r, HV *nginx, SV *sub,
     ngx_str_t **args, ngx_str_t *handler, ngx_str_t *rv)
 {
     SV          *sv;
@@ -718,7 +642,7 @@ ngx_http_perl_call_handler(pTHX_ ngx_http_request_t *r, SV *sub,
 
     PUSHMARK(sp);
 
-    sv = sv_2mortal(sv_bless(newRV_noinc(newSViv(PTR2IV(r))), nginx_stash));
+    sv = sv_2mortal(sv_bless(newRV_noinc(newSViv(PTR2IV(r))), nginx));
     XPUSHs(sv);
 
     if (args) {
@@ -810,6 +734,9 @@ ngx_http_perl_eval_anon_sub(pTHX_ ngx_str_t *handler, SV **sv)
 
     if (ngx_strncmp(p, "sub ", 4) == 0 || ngx_strncmp(p, "use ", 4) == 0) {
         *sv = eval_pv((char *) p, FALSE);
+
+        /* eval_pv() does not set ERRSV on failure */
+
         return;
     }
 
@@ -827,8 +754,6 @@ ngx_http_perl_create_main_conf(ngx_conf_t *cf)
         return NGX_CONF_ERROR;
     }
 
-    pmcf->interp_max = NGX_CONF_UNSET_UINT;
-
     if (ngx_array_init(&pmcf->requires, cf->pool, 1, sizeof(u_char *))
         != NGX_OK)
     {
@@ -844,60 +769,33 @@ ngx_http_perl_init_main_conf(ngx_conf_t *cf, void *conf)
 {
     ngx_http_perl_main_conf_t *pmcf = conf;
 
-#if (NGX_HAVE_PERL_CLONE || NGX_HAVE_PERL_MULTIPLICITY)
-    ngx_conf_init_uint_value(pmcf->interp_max, 10);
-#else
-    ngx_conf_init_uint_value(pmcf->interp_max, 1);
-#endif
-
-    pmcf->free_perls = ngx_pcalloc(cf->pool,
-                                  pmcf->interp_max * sizeof(PerlInterpreter *));
-    if (pmcf->free_perls == NULL) {
-        return NGX_CONF_ERROR;
-    }
-
     if (pmcf->perl == NULL) {
         if (ngx_http_perl_init_interpreter(cf, pmcf) != NGX_CONF_OK) {
             return NGX_CONF_ERROR;
         }
     }
 
-#if !(NGX_HAVE_PERL_CLONE)
-    ngx_http_perl_free_interpreter(pmcf, pmcf->perl);
-#endif
-
     return NGX_CONF_OK;
 }
 
 
-#if (NGX_HAVE_PERL_CLONE || NGX_HAVE_PERL_MULTIPLICITY)
+#if (NGX_HAVE_PERL_MULTIPLICITY)
 
 static void
 ngx_http_perl_cleanup_perl(void *data)
 {
     PerlInterpreter  *perl = data;
 
+    PERL_SET_CONTEXT(perl);
+
     (void) perl_destruct(perl);
 
     perl_free(perl);
-
-    PERL_SYS_TERM();
 }
 
 #endif
 
 
-static void
-ngx_http_perl_cleanup_sv(void *data)
-{
-    ngx_http_perl_cleanup_t  *cln = data;
-
-    dTHXa(cln->perl);
-
-    SvREFCNT_dec(cln->sv);
-}
-
-
 static ngx_int_t
 ngx_http_perl_preconfiguration(ngx_conf_t *cf)
 {
@@ -988,8 +886,6 @@ ngx_http_perl(ngx_conf_t *cf, ngx_command_t *cmd, void *conf)
     ngx_http_perl_loc_conf_t *plcf = conf;
 
     ngx_str_t                  *value;
-    ngx_pool_cleanup_t         *cln;
-    ngx_http_perl_cleanup_t    *pcln;
     ngx_http_core_loc_conf_t   *clcf;
     ngx_http_perl_main_conf_t  *pmcf;
 
@@ -1009,16 +905,12 @@ ngx_http_perl(ngx_conf_t *cf, ngx_command_t *cmd, void *conf)
         }
     }
 
-    cln = ngx_pool_cleanup_add(cf->pool, sizeof(ngx_http_perl_cleanup_t));
-    if (cln == NULL) {
-        return NGX_CONF_ERROR;
-    }
-
     plcf->handler = value[1];
 
     {
 
     dTHXa(pmcf->perl);
+    PERL_SET_CONTEXT(pmcf->perl);
 
     ngx_http_perl_eval_anon_sub(aTHX_ &value[1], &plcf->sub);
 
@@ -1034,11 +926,6 @@ ngx_http_perl(ngx_conf_t *cf, ngx_command_t *cmd, void *conf)
 
     }
 
-    cln->handler = ngx_http_perl_cleanup_sv;
-    pcln = cln->data;
-    pcln->sv = plcf->sub;
-    pcln->perl = pmcf->perl;
-
     clcf = ngx_http_conf_get_module_loc_conf(cf, ngx_http_core_module);
     clcf->handler = ngx_http_perl_handler;
 
@@ -1051,9 +938,7 @@ ngx_http_perl_set(ngx_conf_t *cf, ngx_command_t *cmd, void *conf)
 {
     ngx_int_t                   index;
     ngx_str_t                  *value;
-    ngx_pool_cleanup_t         *cln;
     ngx_http_variable_t        *v;
-    ngx_http_perl_cleanup_t    *pcln;
     ngx_http_perl_variable_t   *pv;
     ngx_http_perl_main_conf_t  *pmcf;
 
@@ -1091,16 +976,12 @@ ngx_http_perl_set(ngx_conf_t *cf, ngx_command_t *cmd, void *conf)
         }
     }
 
-    cln = ngx_pool_cleanup_add(cf->pool, sizeof(ngx_http_perl_cleanup_t));
-    if (cln == NULL) {
-        return NGX_CONF_ERROR;
-    }
-
     pv->handler = value[2];
 
     {
 
     dTHXa(pmcf->perl);
+    PERL_SET_CONTEXT(pmcf->perl);
 
     ngx_http_perl_eval_anon_sub(aTHX_ &value[2], &pv->sub);
 
@@ -1116,11 +997,6 @@ ngx_http_perl_set(ngx_conf_t *cf, ngx_command_t *cmd, void *conf)
 
     }
 
-    cln->handler = ngx_http_perl_cleanup_sv;
-    pcln = cln->data;
-    pcln->sv = pv->sub;
-    pcln->perl = pmcf->perl;
-
     v->get_handler = ngx_http_perl_variable;
     v->data = (uintptr_t) pv;
 
@@ -1128,17 +1004,8 @@ ngx_http_perl_set(ngx_conf_t *cf, ngx_command_t *cmd, void *conf)
 }
 
 
-static char *
-ngx_http_perl_interp_max_unsupported(ngx_conf_t *cf, void *post, void *data)
+static void
+ngx_http_perl_exit(ngx_cycle_t *cycle)
 {
-#if (NGX_HAVE_PERL_CLONE || NGX_HAVE_PERL_MULTIPLICITY)
-
-    return NGX_CONF_OK;
-
-#else
-
-    return "to use perl_interp_max you have to build perl with "
-           "-Dusemultiplicity or -Dusethreads options";
-
-#endif
+    PERL_SYS_TERM();
 }
